/* hoc - higher order calculator
 *
 * Purpose:
 *	Online calculator with full suite of math functions.
 *
 * Credits:
 *	Kernighan and Pike, "The UNIX Programming Environment",
 *	version 3, with the enhancements suggested in the exercises.
 *
 * $Author: jkcohen $
 * $Source: /mnt1/h/jkcohen/Source/C/Hoc/RCS/hoc.y,v $
 * $Revision: 1.1 $ ; $Date: 87/03/16 21:16:25 $
 * $State: Stab $
*/

%{
/*********************** self documentation **********************/
char *sdoc = "\
								\n\
HOC - hoc [-]							\n\
								\n\
hoc is an enhanced version 3 of Kernighan and Pike's calculator	\n\
You exit by typing a ctrl-D.  It works like so:			\n\
hoc								\n\
3 + 4								\n\
	7							\n\
3^4								\n\
	81							\n\
sqrt(#)								\n\
	9							\n\
sin(PI/6)							\n\
	.5							\n\
3e-3 * #							\n\
	0.075							\n\
c = 5687; tmin = 1.2						\n\
c * tmin							\n\
	6824.4							\n\
^D								\n\
# is the symbol for the last value printed.			\n\
functions and constants:					\n\
								\n\
DEG (deg/radian), E,  GAMMA (Euler), PHI (golden), PI		\n\
abs, acos, asin, atan, atan2, ceil, cos, cosh, exp, floor,	\n\
gamma, hypot, int, j0, j1, jn, log, log10, pow, rand, sin,	\n\
sinh, sqrt, srand, tan, tanh, y0, y1, yn			\n\
";
/*****************************************************************/

#include "hoc.h"
#include "su.h"
#include "jc.h"
extern double floor(), Pow();
double lastno; 	/* last printed value */
%}
%union {
	double	val;	/* actual value */
	Symbol	*sym;	/* symbol table pointer */
}
%token	<val>	NUMBER ENDLIST
%token	<sym>	CON VAR BLTIN BLTIN2 BLTIN0 UNDEF
%type 	<val>	expr asgn
%right	'='
%left	'+' '-'
%left	'*' '/' '%'
%left	UNARYMINUS UNARYPLUS
%right	'^'	/* exponentiation */
%%
list:	  /* nothing */
	| list ENDLIST
	| list asgn ENDLIST
	| list expr ENDLIST	{ lastno = $2; printf("\t%.8g\n", $2); }
	| list error ENDLIST	{ yyerrok; }
;
asgn:	  CON '=' expr	{ execerror("can't assign to constant: ", $1->name); }
	| VAR '=' expr	{ $$ = $1->u.val = $3; $1->type = VAR; }
;
expr:	  NUMBER	{ $$ = $1; }
	| CON		{ $$ = $1->u.val; }
	| VAR		{
		If $1->type == UNDEF Do
			execerror("undefined variable: ", $1->name);
  		Endif
		$$ = $1->u.val; }
	| BLTIN '(' expr ')'	{ $$ = (*($1->u.ptr))($3); }
	| BLTIN2 '(' expr ',' expr ')'	{
		$$ = (*($1->u.ptr))($3, $5); }
	| BLTIN0 '(' ')'	{ $$ = (*($1->u.ptr))(); }
	| asgn
	| expr '+' expr { $$ = $1 + $3; }
	| expr '-' expr { $$ = $1 - $3; }
	| expr '*' expr { $$ = $1 * $3; }
	| expr '/' expr {
		If $3 == 0.0 Do
			execerror("division by zero", "");
		Endif
		$$ = $1 / $3; }
	| expr '%' expr {	/* definition as per Knuth, v1 */
		If $3 == 0.0 Do
			$$ = $1;
		Elsedo
			$$ = $1 - $3 * floor($1 / $3);
		Endif }
	| expr '^' expr { $$ = Pow($1, $3); }
	| '(' expr ')'  { $$ = $2 ; }
	| '-' expr %prec UNARYMINUS { $$ = -$2; }
	| '+' expr %prec UNARYMINUS { $$ = $2; }
;
%%
	/* end of grammar */


#include <signal.h>
#include <setjmp.h>
jmp_buf begin;

int	lineno = 0;

main(argc, argv)	/* hoc3 */
	int argc;
	char *argv[];
Begin
	int fpecatch(); /* floating point error catcher */

	/* Set up standard SU args */
	xargc = argc; xargv = argv;

	/* Code begins - check usage */
	If xargc != 1 Do
		selfdoc();
	Endif

	init();
	setjmp(begin);
	signal(SIGFPE, fpecatch);
	yyparse();
End


execerror(s, t)  /* recover from runtime error */
register char *s, *t;
Begin
	warning(s, t);
	longjmp(begin, 0);
End


int fpecatch()	/* catch floating point exceptions */
Begin
	execerror("floating point exception", (char *) 0);
End

int yylex()
Begin
	register int c;

	While (c = getchar()) == ' ' || c == '\t' Do
		;
	Endwhile
	If c == EOF Do  /* archaic yyparse uses 0 for end_of_file */
		Return 0;
	Endif
	If c == '.' || isdigit(c) Do  /* number */
		ungetc(c, stdin);
		scanf("%lf", &yylval.val);
		Return NUMBER;
	Endif
	If c == '#' Do
		yylval.val = lastno;
		Return NUMBER;
	Endif
	If isalpha(c) Do
		Symbol *s;
		char sbuf[100], *p = sbuf;
		Loop
			*p++ = c;
		Exitif( (c = getchar()) == EOF || !isalnum(c) );
		Endloop
		ungetc(c, stdin);
		*p = '\0';
		If (s = lookup(sbuf)) == 0 Do
			s = install(sbuf, UNDEF, 0.0);
		Endif
		yylval.sym = s;
		Return s->type == UNDEF ? VAR : s->type;
	Endif
	If c == '\n' Do
		lineno++;
	Endif
	If c == '\n' || c == ';' Do
		Return ENDLIST;
	Endif
	Return c;
End


yyerror(s)		/* called for yacc syntax error */
register char *s;
Begin
	warning(s, (char *) 0 );
End


warning(s, t)	/* print warning message */
register char *s, *t;
Begin
	fprintf(stderr, "%s: %s", xargv[0], s);
	If t Do
		fprintf(stderr, "%s", t);
	Endif
	fprintf(stderr, " near line %d\n", lineno);
End
